@@ -54,7 +54,7 @@ class MessageOutput:
     def setFilename(self, filename):
         self.filename = filename
 
-    def outputMessage(self, text, lineno = 0, comment = None, spacepreserve = 0):
+    def outputMessage(self, text, lineno = 0, comment = None, spacepreserve = 0, tag = None):
         """Adds a string to the list of messages."""
         if (text.strip() != ''):
             t = escapePoString(normalizeString(text, not spacepreserve))
@@ -67,26 +67,43 @@ class MessageOutput:
                 if spacepreserve:
                     self.nowrap[t] = 1
                 if t in self.linenos.keys():
-                    self.linenos[t].append((self.filename, lineno))
+                    self.linenos[t].append((self.filename, tag, lineno))
                 else:
-                    self.linenos[t] = [ (self.filename, lineno) ]
+                    self.linenos[t] = [ (self.filename, tag, lineno) ]
                 if (not self.do_translations) and comment and not t in self.comments:
                     self.comments[t] = comment
             else:
                 if t in self.linenos.keys():
-                    self.linenos[t].append((self.filename, lineno))
+                    self.linenos[t].append((self.filename, tag, lineno))
                 else:
-                    self.linenos[t] = [ (self.filename, lineno) ]
+                    self.linenos[t] = [ (self.filename, tag, lineno) ]
                 if comment and not t in self.comments:
                     self.comments[t] = comment
 
+    def outputHeader(self, out):
+        import time
+        out.write("""msgid ""
+msgstr ""
+"Project-Id-Version: PACKAGE VERSION\\n"
+"POT-Creation-Date: %s\\n"
+"PO-Revision-Date: YEAR-MO-DA HO:MI+ZONE\\n"
+"Last-Translator: FULL NAME <EMAIL@ADDRESS>\\n"
+"Language-Team: LANGUAGE <LL@li.org>\\n"
+"MIME-Version: 1.0\\n"
+"Content-Type: text/plain; charset=UTF-8\\n"
+"Content-Transfer-Encoding: 8bit\\n"
+
+""" % (time.strftime("%Y-%m-%d %H:%M%z")))
+
     def outputAll(self, out):
+        self.outputHeader(out)
+        
         for k in self.messages:
             if k in self.comments:
                 out.write("#. %s\n" % (self.comments[k].replace("\n","\n#. ")))
             references = ""
             for reference in self.linenos[k]:
-                references += "%s:%d " % (reference[0], reference[1])
+                references += "%s:%d(%s) " % (reference[0], reference[2], reference[1])
             out.write("#: %s\n" % (references))
             if k in self.nowrap and self.nowrap[k]:
                 out.write("#, no-wrap\n")
@@ -99,7 +116,6 @@ class MessageOutput:
 
 
 def normalizeNode(node):
-    #print node.content
     if not node:
         return
     elif isSpacePreserveNode(node):
@@ -110,7 +126,7 @@ def normalizeNode(node):
         else:
             node.setContent(re.sub('\s+',' ', node.content))
 
-    elif node.children:
+    elif node.children and node.type == 'element':
         child = node.children
         while child:
             normalizeNode(child)
@@ -125,16 +141,17 @@ def normalizeString(text, ignorewhitespace = 1):
     try:
         # Lets add document DTD so entities are resolved
         dtd = doc.intSubset()
-        tmp = ''
-        if expand_entities: # FIXME: we get a "Segmentation fault" in libxml2.parseMemory() when we include DTD otherwise
-            tmp = dtd.serialize()
+        tmp = dtd.serialize('utf-8')
         tmp = tmp + '<norm>%s</norm>' % text
     except:
         tmp = '<norm>%s</norm>' % text
 
     try:
-        #libxml2.replaceEntities(0)
-        tree = libxml2.parseMemory(tmp,len(tmp))
+        ctxt = libxml2.createDocParserCtxt(tmp)
+        if expand_entities:
+            ctxt.replaceEntities(1)
+        ctxt.parseDocument()
+        tree = ctxt.doc()
         newnode = tree.getRootElement()
     except:
         print >> sys.stderr, """Error while normalizing string as XML:\n"%s"\n""" % (text)
@@ -155,11 +172,11 @@ def normalizeString(text, ignorewhitespace = 1):
 
 def stringForEntity(node):
     """Replaces entities in the node."""
-    text = node.serialize()
+    text = node.serialize('utf-8')
     try:
         # Lets add document DTD so entities are resolved
         dtd = node.doc.intSubset()
-        tmp = dtd.serialize() + '<norm>%s</norm>' % text
+        tmp = dtd.serialize('utf-8') + '<norm>%s</norm>' % text
         next = 1
     except:
         tmp = '<norm>%s</norm>' % text
@@ -193,8 +210,8 @@ def unEscapePoString(text):
 def getTranslation(text, spacepreserve = 0):
     """Returns a translation via gettext for specified snippet.
 
-    text may be a string when it is used verbatim, or a tuple (pair), with
-    first component being a string, and the other being a list of replacements.
+    text should be a string to look for, spacepreserve set to 1
+    when spaces should be preserved.
     """
     text = normalizeString(text, not spacepreserve)
     if (text.strip() == ''):
@@ -203,7 +220,7 @@ def getTranslation(text, spacepreserve = 0):
     if file:
         gt = gettext.GNUTranslations(file)
         if gt:
-            return gt.gettext(text)
+            return gt.ugettext(text.decode('utf-8'))
     return text
 
 def startTagForNode(node):
@@ -216,12 +233,16 @@ def startTagForNode(node):
         for p in node.properties:
             if p.type == 'attribute':
                 # FIXME: This part sucks
-                try:
-                    params += ' %s:%s="%s"' % (p.ns().name, p.name, p.content)
-                except:
-                    params += ' %s="%s"' % (p.name, p.content)
+                params += p.serialize('utf-8')
     return result+params
         
+def endTagForNode(node):
+    if not node:
+        return 0
+
+    result = node.name
+    return result
+        
 def isFinalNode(node):
     if automatic:
         auto = autoNodeIsFinal(node)
@@ -235,6 +256,15 @@ def isFinalNode(node):
     #node.type =='text' or not node.children or
     if node.type == 'element' and node.name in ultimate_tags:
         return 1
+    elif node.children:
+        final_children = 1
+        child = node.children
+        while child and final_children:
+            if not isFinalNode(child):
+                final_children = 0
+            child = child.next
+        if final_children:
+            return 1
     return 0
 
 def ignoreNode(node):
@@ -272,42 +302,46 @@ def getCommentForNode(node):
     else:
         return None
 
+
 def replaceNodeContentsWithText(node,text):
     """Replaces all subnodes of a node with contents of text treated as XML."""
-    #print >> sys.stderr, text
     if node.children:
-        tmp = '<%s>%s</%s>' % (startTagForNode(node), text, node.name)
+        starttag = node.name #startTagForNode(node)
+        endtag = endTagForNode(node)
         try:
-            newnode = libxml2.parseMemory(tmp,len(tmp))
+            # Lets add document DTD so entities are resolved
+            dtd = doc.intSubset()
+            tmp = ''
+            if expand_entities: # FIXME: we get a "Segmentation fault" in libxml2.parseMemory() when we include DTD otherwise
+                tmp = dtd.serialize('utf-8')
+            tmp = tmp + '<%s>%s</%s>' % (starttag, text, endtag)
         except:
-            print >> sys.stderr, """Error while parsing translation as XML:\n"%s"\n""" % (text)
+            tmp = '<%s>%s</%s>' % (starttag, text, endtag)
+
+        try:
+            ctxt = libxml2.createDocParserCtxt(tmp.encode('utf-8'))
+            ctxt.replaceEntities(0)
+            ctxt.parseDocument()
+            newnode = ctxt.doc()
+        except:
+            print >> sys.stderr, """Error while parsing translation as XML:\n"%s"\n""" % (text.encode('utf-8'))
             return
-        if newnode.children and newnode.children.children:
+
+        newelem = newnode.getRootElement()
+        if newelem and newelem.children:
             free = node.children
             while free:
                 next = free.next
                 free.unlinkNode()
                 free = next
-            node.addChildList(newnode.children.children)
+
+            node.addChildList(newelem.children)
         else:
             # In practice, this happens with tags such as "<para>    </para>" (only whitespace in between)
             pass
     else:
         node.setContent(text)
 
-def processFinalTag(node, outtxt):
-    """node must be isFinalTag, this must be checked before calling this function."""
-    global semitrans
-    if mode == 'merge':
-        outtxt = getTranslation(outtxt, isSpacePreserveNode(node))
-        for i in semitrans.keys():
-            outtxt = outtxt.replace('<placeholder-%d/>' % (i), semitrans[i])
-        replaceNodeContentsWithText(node,outtxt)
-    else:
-        if node.name not in ignored_tags:
-            msg.outputMessage(outtxt, node.lineNo(), getCommentForNode(node), isSpacePreserveNode(node))
-    return outtxt
-
 def autoNodeIsFinal(node):
     """Returns 1 if node is text node, contains non-whitespace text nodes or entities."""
     final = 0
@@ -331,8 +365,11 @@ def worthOutputting(node):
     """
     worth = 1
     parent = node.parent
-    while parent:
+    final = isFinalNode(node) and node.name not in ignored_tags
+    while not final and parent:
         if isFinalNode(parent):
+            final = 1 # reset if we've got to one final tag
+        if final and (parent.name not in ignored_tags) and worthOutputting(parent):
             worth = 0
             break
         parent = parent.parent
@@ -341,43 +378,52 @@ def worthOutputting(node):
 
     return autoNodeIsFinal(node)
     
-def processElementTag(node):
+def processElementTag(node, replacements, restart = 0):
     """Process node with node.type == 'element'."""
     if node.type == 'element':
-        global PlaceHolder
-        global semitrans
-        final = isFinalNode(node)
         outtxt = ''
-        if final:
-            storeholder = PlaceHolder
-            PlaceHolder = 0
-            storesemi = semitrans
-            semitrans = {}
+        if restart:
+            myrepl = []
+        else:
+            myrepl = replacements
+
+        submsgs = []
+
         child = node.children
         while child:
-            if isFinalNode(child):
-                PlaceHolder += 1
-                newmsg = doSerialize(child)
-                result = '<placeholder-%d/>' % (PlaceHolder)
-                if mode=='merge':
-                    semitrans[PlaceHolder] = getTranslation(newmsg, isSpacePreserveNode(node))
+            if (isFinalNode(child)) or (child.type == 'element' and worthOutputting(child)):
+                myrepl.append(processElementTag(child, myrepl, 1))
+                outtxt += '<placeholder-%d/>' % (len(myrepl))
             else:
-                result = doSerialize(child)
-            outtxt += result
+                if child.type == 'element':
+                    (starttag, content, endtag, translation) = processElementTag(child, myrepl, 0)
+                    outtxt += '<%s>%s</%s>' % (starttag, content, endtag)
+                else:
+                    outtxt += doSerialize(child)
+
             child = child.next
 
-        if final:
-            outtxt = processFinalTag(node, outtxt)
-            PlaceHolder = storeholder
-            semitrans = storesemi
-            return '<%s>%s</%s>' % (startTagForNode(node), outtxt, node.name)
+        if mode == 'merge':
+            translation = getTranslation(outtxt, isSpacePreserveNode(node))
         else:
+            translation = outtxt
+        starttag = startTagForNode(node)
+        endtag = endTagForNode(node)
+
+        if restart or worthOutputting(node):
+            i = 0
+            while i < len(myrepl):
+                replacement = '<%s>%s</%s>' % (myrepl[i][0], myrepl[i][3], myrepl[i][2])
+                i += 1
+                translation = translation.replace('<placeholder-%d/>' % (i), replacement)
+
             if worthOutputting(node):
                 if mode == 'merge':
-                    replaceNodeContentsWithText(node,getTranslation(outtxt, isSpacePreserveNode(node)))
+                    replaceNodeContentsWithText(node, translation)
                 else:
-                    msg.outputMessage(outtxt, node.lineNo(), getCommentForNode(node), isSpacePreserveNode(node))
-            return '<%s>%s</%s>' % (startTagForNode(node), outtxt, node.name)
+                    msg.outputMessage(outtxt, node.lineNo(), getCommentForNode(node), isSpacePreserveNode(node), tag = node.name)
+
+        return (starttag, outtxt, endtag, translation)
     else:
         raise Exception("You must pass node with node.type=='element'.")
 
@@ -426,7 +472,9 @@ def doSerialize(node):
     elif node.type == 'text':
         return node.serialize('utf-8')
     elif node.type == 'element':
-        return processElementTag(node)
+        repl = []
+        (starttag, content, endtag, translation) = processElementTag(node, repl, 1)
+        return '<%s>%s</%s>' % (starttag, content, endtag)
     else:
         child = node.children
         outtxt = ''
@@ -521,9 +569,10 @@ libxml2.registerErrorHandler(xml_error_handler, None)
 
 
 # Main program start
+if __name__ != '__main__': raise NotImplementedError
 
 # Parameters
-submodes_path = "/home/danilo/cvs/i18n/xml2po/modes"
+submodes_path = "/home/danilo/cvs/gnom/gnome-doc-utils/xml2po/modes"
 default_mode = 'docbook'
 
 filename = ''
@@ -542,10 +591,46 @@ output  = '-' # this means to stdout
 
 import getopt, fileinput
 
+def usage (with_help = False):
+        print >> sys.stderr, "Usage:  %s [OPTIONS] [XMLFILE]..." % (sys.argv[0])
+	if (with_help):
+        	print >> sys.stderr, """
+OPTIONS may be some of:
+    -a    --automatic-tags     Automatically decides if tags are to be considered
+                                 "final" or not
+    -k    --keep-entities      Don't expand entities
+    -e    --expand-all-entities  Expand ALL entities (including SYSTEM ones)
+    -m    --mode=TYPE          Treat tags as type TYPE (default: docbook)
+    -o    --output=FILE        Print resulting text (XML or POT) to FILE
+    -p    --po-file=FILE       Specify PO file containing translation, and merge
+                                 Overwrites temporary file .xml2po.mo.
+    -r    --reuse=FILE         Specify translated XML file with the same structure
+    -t    --translation=FILE   Specify MO file containing translation, and merge
+    -u    --update-translation=LANG.po   Updates a PO file using msgmerge program
+    -v    --version            Output version of the xml2po program
+
+    -h    --help               Output this message
+
+EXAMPLES:
+    To create a POTemplate book.pot from input files chapter1.xml and
+    chapter2.xml, run the following:
+        %s -o book.pot chapter1.xml chapter2.xml
+
+    After translating book.pot into de.po, merge the translations back,
+    using -p option for each XML file:
+        %s -p de.po chapter1.xml > chapter1.de.xml
+        %s -p de.po chapter2.xml > chapter2.de.xml
+""" % (sys.argv[0], sys.argv[0], sys.argv[0])
+        sys.exit(0)
+
+if len(sys.argv) < 2: usage()
+
 args = sys.argv[1:]
-opts, args = getopt.getopt(args, 'avhkem:t:o:p:u:r:',
+try: opts, args = getopt.getopt(args, 'avhkem:t:o:p:u:r:',
                            ['automatic-tags','version', 'help', 'keep-entities', 'expand-all-entities', 'mode=', 'translation=',
                             'output=', 'po-file=', 'update-translation=', 'reuse=' ])
+except getopt.GetoptError: usage(True)
+
 for opt, arg in opts:
     if opt in ('-m', '--mode'):
         default_mode = arg
@@ -575,35 +660,7 @@ for opt, arg in opts:
         print VERSION
         sys.exit(0)
     elif opt in ('-h', '--help'):
-        print >> sys.stderr, "Usage:  %s [OPTIONS] [XMLFILE]..." % (sys.argv[0])
-        print >> sys.stderr, """
-OPTIONS may be some of:
-    -a    --automatic-tags     Automatically decides if tags are to be considered
-                                 "final" or not (overrides -f and -i options)
-    -k    --keep-entities      Don't expand entities
-    -e    --expand-all-entities  Expand ALL entities (including SYSTEM ones)
-    -m    --mode=TYPE          Treat tags as type TYPE (default: docbook)
-    -o    --output=FILE        Print resulting text (XML or POT) to FILE
-    -p    --po-file=FILE       Specify PO file containing translation, and merge
-                                 Overwrites temporary file .xml2po.mo.
-    -r    --reuse=FILE         Specify translated XML file with the same structure
-    -t    --translation=FILE   Specify MO file containing translation, and merge
-    -u    --update-translation=LANG.po   Updates a PO file using msgmerge program
-    -v    --version            Output version of the xml2po program
-
-    -h    --help               Output this message
-
-EXAMPLES:
-    To create a POTemplate book.pot from input files chapter1.xml and
-    chapter2.xml, run the following:
-        %s -o book.pot chapter1.xml chapter2.xml
-
-    After translating book.pot into de.po, merge the translations back,
-    using -p option for each XML file:
-        %s -p de.po chapter1.xml > chapter1.de.xml
-        %s -p de.po chapter2.xml > chapter2.de.xml
-""" % (sys.argv[0], sys.argv[0], sys.argv[0])
-        sys.exit(0)
+    	usage(True)
 
 # Treat remaining arguments as XML files
 while args:
